/*
	thornberry/frequency/FrequencyEditor.js

	Author: wamesbury
	Created: *Unknown*
	Purpose: Discipline frequency editor
	History:
	2022-03-29 DJA ND-11859 Consider new AllowVisitDurationTiming property on Discipline Table when determining whether the Visit Length and Visit Timing fields should be available to users
	2018-12-27 DJA ND-8145 ClearFrequency() now also clears the value for cert; Provides better support for rolling back the cert value in DisciplineEditor when frequency modification is unavailable
*/
define([
	"dojo/_base/declare",
	"thornberry/object/ObjectEditor",
	"thornberry/object/ObjectModel",
	"thornberry/serverRequest",
	"thornberry/frequency/FrequencyCalendar",
	"thornberry/object/TableManager",
	"dojo/_base/array",
	"dojo/_base/lang",
	"dojo/aspect",
	"dojo/store/Memory",
	"dojo/store/Observable",
	"dojo/date",
	"dojo/date/locale",
	"dojo/dom-construct",
	"dojo/dom-style",
	"dojo/dom-class",
	"dojo/query",
	"dojo/Deferred",
	"dijit/form/ValidationTextBox",
	"dijit/form/Textarea"
], function(declare, ObjectEditor, ObjectModel, serverRequest, FrequencyCalendar, TableManager, arrayUtils, lang, aspect, Memory, Observable,
			date, locale, domConstruct, domStyle, domClass, query, Deferred, ValidationTextBox, Textarea) {

		// Requirements:
		// 	ndocNumber variable defined
		// 	frequencySettings object defined										Patient.Frequency.Utility::EditorSettings()
		// 	TableManager.get("Tables.Discipline") exists							Tables.Discipline::RetrieveAsJSON(1)	TODO: Retrieve in-page if missing
		// 	CSS class:
		//		 .claro .dijitCalendarDateTemplate { font-weight: normal; }			TODO: See if this can be removed; might affect other calendar widgets on page
		var numberRegex = "[1-9]([0-9])?";
		var durationRegex = "(0|[1-9]\\d?)?(\\.\\d{1,2})?";		// 1-2 integers with 0-2 decimal points, optionally with a leading zero

		var withinPerStore = new Memory({data:
			[
				{id: "per", name: "per"},
				{id: "within", name: "within"}
			]
		});

		var withinTimeUnitStore = new Memory({data:
			[
				{ id: "D", name: "days" },
				{ id: "W", name: "weeks" },
				{ id: "M", name: "months" }
			]
		});

		var perTimeUnitStore = new Memory({data:
			[
				{id: "D",	 name: "day", 			plural: "days"},
				{id: "W",	 name: "week", 			plural: "weeks"},
				{id: "M",	 name: "month", 		plural: "months"},
				{id: "QOD",	 name: "q other day", 	plural: "days"},
				{id: "QOW",	 name: "q other week", 	plural: "weeks"},
				{id: "Q60D", name: "q 60 days",		plural: ""}
			]
		});

		var visitTimingStore = new Memory({data:
			[
				{id: "NA", name: "N/A"},
				{id: "M", name: "morning"},
				{id: "A", name: "afternoon"},
				{id: "E", name: "evening"},
				{id: "N", name: "night"}
			]
		});

		frequencyItemProperties = [
			{ name: "startDate",		type: "date", required: true },
			{ name: "stopDate",			type: "date", fieldProperties: {disabled: true} },

			{ name: "visitNumMin",		type: "text", required: true, fieldProperties: {style: "width: 45px", regExp: numberRegex} },
			{ name: "visitNumMax",		type: "text", required: true, fieldProperties: {style: "width: 45px", regExp: numberRegex} },

			{ name: "withinPer",		type: "tableItem", required: true, table: withinPerStore},

			{ name: "withinTimeUnit",	type: "tableItem", table: withinTimeUnitStore },
			{ name: "withinQuantity",	type: "text", fieldProperties: {style: "width: 45px", regExp: numberRegex}},

			{ name: "perTimeUnit",		type: "tableItem", table: perTimeUnitStore },
			{ name: "perQuantity",		type: "text", fieldProperties: {style: "width: 45px", regExp: numberRegex} },

			{ name: "visitDurationMin", type: "text", fieldProperties: {style: "width: 45px", regExp: durationRegex} },
			{ name: "visitDurationMax", type: "text", fieldProperties: {style: "width: 45px", regExp: durationRegex} },
			{ name: "visitTiming",		type: "tableItem", table: visitTimingStore },

			{ name: "breakReason",		type: "text", fieldProperties: {style: "width: 400px" } }
		];

		var frequencyItemModel = new ObjectModel({propertyDefinitions: frequencyItemProperties});
		var FrequencyItem = frequencyItemModel.getObject();

		var FrequencyItemEditor = declare([ObjectEditor], {
			objectModel: frequencyItemModel,
			frequencyEditor: null,
			firstEditor: false,
			lastEditor: false,
			breakStart: null,
			breakStop: null,

			templateString:
				'<div class="${baseClass}" data-dojo-attach-point="containerNode" style="width:100%;height:100%;border-top:1px solid black;border-bottom:1px solid black;margin-left:5px;">'+
					'<div class="content-area" data-dojo-attach-point="breakReasonContainer" style="border-left:20px solid #FFCCCC; border-bottom: 1px solid black; display: none; padding: 0px;">'+
						'<div class="content-flex-container">'+
							'<div class="flex-field-container">'+
								'<div class="field-name-larger">'+
									'<label style="font-weight: bold;">Break in Frequency!</label>'+
								'</div>'+
								'<div class="field-value field-value-medium">'+
									'<label data-dojo-attach-point="breakDurationLabel"></label>'+
								'</div>'+
							'</div>'+
						'</div>'+
						'<div class="content-flex-container">'+
							'<div class="flex-field-container">'+
								'<div class="field-name field-name-small">'+
									'<label>Reason</label>'+
								'</div>'+
								'<div class="field-value field-value-medium">'+
									'<input data-dojo-attach-point="breakReasonNode">'+
								'</div>'+
							'</div>'+
						'</div>'+
					'</div>'+
					'<div class="content-area" data-dojo-attach-point="frequencyContainer" style="border-left:20px solid; padding: 0px;">'+
						'<div class="content-flex-container">'+
							'<div class="flex-field-container vertical-field-align" style="flex-direction: column;">'+
								'<div class="field-name field-name-smallest">'+
									'<label>Start</label>'+
								'</div>'+
								'<div class="field-value field-value-medium-small">'+
									'<input data-dojo-attach-point="startDateNode">'+
								'</div>'+
							'</div>'+
							'<div class="flex-field-container vertical-field-align" style="flex-direction: column;">'+
								'<div class="field-name field-name-smallest">'+
									'<label>Stop</label>'+
								'</div>'+
								'<div class="field-value field-value-medium-small">'+
									'<span> </span>'+
									'<label data-dojo-attach-point="stopDateLabel"></label>'+
								'</div>'+
							'</div>'+
						'</div>'+
						'<div class="content-flex-container">'+
							'<div class="flex-field-container">'+
								'<div class="field-name field-name-small"></div>'+
							'</div>'+
							'<div class="flex-field-container vertical-field-align" style="flex-direction: column;">'+
								'<div class="field-name field-name-small">'+
									'<label>Visits</label>'+
								'</div>'+
								'<div class="field-value field-value-large" style="display: inline;">'+
									'<input data-dojo-attach-point="visitNumMinNode">'+
									'<label> to </label>'+
									'<input data-dojo-attach-point="visitNumMaxNode">'+
								'</div>'+
							'</div>'+
							'<div class="flex-field-container vertical-field-align">'+
								'<div class="field-name field-name-smallest"></div>'+
							'</div>'+
							'<div class="flex-field-container vertical-field-align" style="flex-direction: column;">'+
								'<div class="field-name field-name-small">'+
									'<label>Frequency</label>'+
								'</div>'+
								'<div class="field-value field-value-large" style="display: inline;">'+
									'<input data-dojo-attach-point="withinPerNode">'+
									'<span> </span>'+
									'<span data-dojo-attach-point="withinFieldsContainer">'+
										'<input data-dojo-attach-point="withinQuantityNode">'+
										'<span> </span>'+
										'<input data-dojo-attach-point="withinTimeUnitNode">'+
									'</span>'+
									'<span data-dojo-attach-point="perFieldsContainer">'+
										'<input data-dojo-attach-point="perTimeUnitNode">'+
										'<label> for </label>'+
										'<input data-dojo-attach-point="perQuantityNode">'+
										'<label data-dojo-attach-point="perQuantityLabel"></label>'+
									'</span>'+
								'</div>'+
							'</div>'+
						'</div>'+
						'<div class="content-flex-container">'+
							'<div class="flex-field-container">'+
								'<div class="field-name field-name-medium"></div>'+
							'</div>'+
							'<div class="flex-field-container vertical-field-align" style="flex-direction: column;">'+
								'<div class="field-name field-name-medium">'+
									'<label>Visit Length</label>'+
								'</div>'+
								'<div class="field-value field-value-large" style="display: inline;">'+
									'<input data-dojo-attach-point="visitDurationMinNode">'+
									'<label> to </label>'+
									'<input data-dojo-attach-point="visitDurationMaxNode">'+
									'<label> hrs/visit </label>'+
								'</div>'+
							'</div>'+
							'<div class="flex-field-container vertical-field-align" style="flex-direction: column;">'+
								'<div class="field-name field-name-smallest"></div>'+
							'</div>'+
							'<div class="flex-field-container vertical-field-align" style="flex-direction: column;">'+
								'<div class="field-name field-name-medium">'+
									'<label>Visit Timing</label>'+
								'</div>'+
								'<div class="field-value field-value-large" style="display: inline;">'+
									'<input data-dojo-attach-point="visitTimingNode">'+
								'</div>'+
							'</div>'+
						'</div>'+
					'</div>'+
				'</div>',

			postCreate: function() {
				var t = this;
				t.inherited(arguments);

				t.visitNumMinField.validator = function(value, constraints) {
					return ValidationTextBox.prototype.validator.call(this, value, constraints) && (!constraints.max || (+value <= constraints.max));
				};

				t.visitDurationMinField.validator = function(value, constraints) {
					return ValidationTextBox.prototype.validator.call(this, value, constraints) && (!constraints.max || (+value <= constraints.max));
				};

				t.own(aspect.before(t, "_setObjectAttr", function() {
					if (t.frequencyEditor._frequencyEditorStore.data.length === 0) {
						t.firstEditor = true;
					}
				}));

				t.own(aspect.after(t, "_setObjectAttr", function(object) {
					// Disable visit length fields for unapplicable disciplines
					if (
						t.frequencyEditor.discipline
						&&
						TableManager.get("Tables.Discipline").get(t.frequencyEditor.discipline).AllowVisitDurationTiming == 1
					) {
						t.require("visitDurationMax", (arrayUtils.indexOf(frequencySettings.aideHoursRequired, t.frequencyEditor.discipline) !== -1));
						t.updateFieldPinking();
						t.disable("visitDurationMin", false);
						t.disable("visitDurationMax", false);
						t.disable("visitTiming", false);
					} else {
						t.disable("visitDurationMin", true);
						t.disable("visitDurationMax", true);
						t.disable("visitTiming", true);
					}

					if (typeof object.withinPer == "undefined" && t.firstEditor) {
						t.object.set("withinPer", "per");
					}

					t._postModify();
				}, true));

				t.own(aspect.after(t, "_modifyValue", function() {
					t.frequencyEditor._modifyValue();
				}));

				t.own(aspect.after(t,"_setActionTypeAttr",function(actionType){
					t._set('actionType', actionType);
					if (actionType == "LimitedEdit") {
						t.disable("visitNumMin", true);
						t.disable("visitNumMax", true);
						t.disable("withinPer", true);
						t.disable("withinTimeUnit", true);
						t.disable("withinQuantity", true);
						t.disable("perTimeUnit", true);
						t.disable("perQuantity", true);
						t.disable("visitDurationMin", true);
						t.disable("visitDurationMax", true);
						t.disable("visitTiming", true);
						t.disable("breakReason", true);
					}
				}, true));

				t.registerChangeLogic("visitNumMin", function(value) {
					t.require("visitNumMax", !value);
				});
				t.registerChangeLogic("visitNumMax", function(value) {
					t.require("visitNumMin", !value);
				});

				t.registerChangeLogic("withinPer", function(value) {
					domStyle.set(t.perFieldsContainer, "display", (value !== "per" ? "none" : ""));
					domStyle.set(t.withinFieldsContainer, "display", (value !== "within" ? "none" : ""));

					t.require("perTimeUnit", (value === "per"));
					t.require("perQuantity", (value === "per"));
					t.require("withinQuantity", (value === "within"));
					t.require("withinTimeUnit", (value === "within"));

					if (!t.inObjectLoad) {
						t.object.set("perTimeUnit", "");
						t.object.set("perQuantity", "");
						t.object.set("withinQuantity", "");
						t.object.set("withinTimeUnit", "");
					}

					var lastEditor = t.frequencyEditor.getLastEditor();
					var lastTimeUnit;
					if ((t.object.get("withinPer") == "per") && (t.object.get("perTimeUnit") == "")) {
						lastTimeUnit = lastEditor && lastEditor.object.get("perTimeUnit") || "W";
						t.object.set("perTimeUnit", lastTimeUnit);
					} else if ((t.object.get("withinPer") == "within") && (t.object.get("withinTimeUnit") == "")) {
						lastTimeUnit = lastEditor && lastEditor.object.get("withinTimeUnit") || "W";
						t.object.set("withinTimeUnit", lastTimeUnit);
					}
				});

				t.registerChangeLogic("perTimeUnit", function(value) {
					if (t.object.withinPer === "per") {
						var timeUnit = perTimeUnitStore.get(value);
						if (timeUnit && timeUnit.plural) {
							timeUnit = "&nbsp;" + timeUnit.plural;
						} else {
							timeUnit = "";
						}
						t.perQuantityLabel.innerHTML = timeUnit;

						if (value === "Q60D") {
							t.object.set("perQuantity", "");
							t.disable("perQuantity");
						} else {
							t.disable("perQuantity", false);
						}
					}
				});

				t.registerChangeLogic("startDate", function(value) {
					if (!t.inObjectLoad && !value && t.frequencyEditor.certStart) {
						t.object.set("startDate", t.frequencyEditor.certStart);
					}
				});

				t.registerChangeLogic("stopDate", function(value) {
					var label;
					if (value) {
						var startDate = locale.parse(t.object.startDate, {selector: "date", datePattern: "yyyy-MM-dd"});
						var days = date.difference(startDate, value, "day") + 1;

						label = locale.format(value, {selector: 'date'}) + " (" + days + " " + (days == 1 ? "day" : "days") + ")";
					} else {
						label = "";
					}

					t.stopDateLabel.innerHTML = label;
				});
			},

			_startDateOnValidate: function(value) {
				var t = this;
				var valid = true;
				if (value && t.frequencyEditor.certStart && value < t.frequencyEditor.certStart) {
					actionShowMessage("Frequency cannot be prior to the Cert Period start date.");
					valid = false;
				} else {
					var minEditDate = t.frequencyEditor.get('minEditDate');
					if (value && minEditDate && value < minEditDate) {
						actionShowMessage("You are unable to modify the frequency prior to the Order Start date.");
						valid = false;
					}
				}
				return valid;
			},

			updateConstraints: function() {
				var t = this;
				t.startDateField.set("constraints", {min: t.frequencyEditor.certStart, max: t.frequencyEditor.certEnd});
				t.visitNumMinField.set("constraints", {max: t.object.visitNumMax});
				t.visitDurationMinField.set("constraints", {max: t.object.visitDurationMax});
			},

			updateStatus: function() {
				var t = this;

				if ((t.get("actionType") === "View") || (t.get("actionType") === "LimitedEdit") || ((t.get("required") != false) && (t.isValid()) && (query(".pinkBackground", t.domNode).length == 0))) {
					// Valid and all required fields entered
					t.set("status", "complete");
				} else {
					// Invalid or required fields missing
					t.set("status", "incomplete");
				}
			},

			isValid: function() {
				var t = this;
				for (var property in t.objectModel.propertyDefinitions) {
					var fieldName = t.objectModel.propertyDefinitions[property].name;

					var widget = t[fieldName+"Field"];
					if (widget && (typeof widget.isValid !== "undefined") && !widget.isValid()) {
						return false;
					}
				}
				var valid = true;

				// Need at least min or max, AND a quantity value
				valid = (!!t.object.visitNumMax || !!t.object.visitNumMin) && (t.object.get("perTimeUnit") == "Q60D" || !!t.object.perQuantity || !!t.object.withinQuantity);

				return valid;
			},

			validate: function() {
				var t = this;
				for (var property in t.objectModel.propertyDefinitions) {
					var fieldName = t.objectModel.propertyDefinitions[property].name;

					var widget = t[fieldName + "Field"];
					if (widget && typeof widget.validate !== "undefined") {
						widget.validate();
					}
				}
			},

			_postModify: function () {
				var t = this;
				var setRequired = function() {

				};

				t.updateConstraints();
				t.updateStatus();

				if (!t.frequencyEditor.initialLoad) {
					t.frequencyEditor.getSpans(); //.then(setRequired);
				} else {
					t.setRequired();
				}
				t.inherited(arguments);
			},

			setRequired: function() {
				var t = this;

				var required = false;
				var isFirstEditor = !!t.get("firstEditor");
				var hasPRN = false;
				if ((t.frequencyEditor.object.get("prnVisits") > 0) && (t.frequencyEditor.object.get("prnReason") != "")) {
					hasPRN = true;
				}

				if (isFirstEditor && (!hasPRN)) {
					required = true;
				}
				if (!isFirstEditor && (!t.get("lastEditor"))) {
					required = true;
				}
				if (!!t.object.visitNumMax || !!t.object.visitNumMin || t.object.get("perTimeUnit") != "Q60D" && (!!t.object.perQuantity || !!t.object.withinQuantity)) {
					required = true;
				}

				t.set("required", required);
			},

			_setRequiredAttr: function(required) {
				var t = this;
				t.inherited(arguments);
				t.required = (typeof required === "undefined" ? true : required);
				t.updateStatus();
				t.updateFieldPinking();
			},

			// Modified to let the widget-level 'required' attribute override
			updateFieldPinking: function () {
				var t = this;
				for (var propertyName in t._requiredPropertyIndex) {
					var widget = t[propertyName + "Field"];
					if (widget) {
						var value = widget.get("value");
						var node = (widget.reqBkgdNode ? widget.reqBkgdNode : widget.domNode);
						if ((t.get("required") != false) && t._requiredPropertyIndex[propertyName] && !widget.get("disabled") && (value === "" || /^\s+$/.test(value) || /*NaN check*/ value != value || value === null || (typeof value === "object" && value.length === 0))) {
							domClass.add(node, "pinkBackground");
						} else {
							domClass.remove(node, "pinkBackground");
						}
					}
				}
			},

			_setLastEditorAttr: function (lastEditor) {
				var t = this;
				t.lastEditor = lastEditor;
				t.updateFieldPinking();
			},

			_setStatusAttr: function(value) {
				var t = this;
				t.status = value;
				var highlightColor = null;
				switch(value) {
					case "complete":
						highlightColor = "#99FF99"; // rgb(153,255,153)
						break;
					case "incomplete":
						highlightColor = "#FFFF99";	// rgb(255,255,153)
						break;
					default:
						highlightColor = "#99FF99";
				}

				domStyle.set(t.frequencyContainer, "border-left-color", highlightColor);
				this.highlightColor = highlightColor;
			},

			containsDate: function(value) {
				var t = this;
				var frequencyStart = locale.parse(t.object.startDate, {selector: "date", datePattern: "yyyy-MM-dd"});
				var frequencyStop = locale.parse(t.object.stopDate, {selector: "date", datePattern: "yyyy-MM-dd"});
				if ((date.compare(value, frequencyStart, "date") >= 0) && (date.compare(value, frequencyStop, "date") <= 0)) {
					return true;
				}
				return false;
			},

			breakContainsDate: function(value) {
				var t = this;
				if ((date.compare(value, t.breakStart, "date") >= 0) && (date.compare(value, t.breakStop, "date") <= 0)) {
					return true;
				}
				return false;
			}
		});


		frequencyProperties = [
			{ name: "prnVisits", required: false, type: "text", fieldProperties: {style: "width: 45px", regExp: numberRegex}},
			{ name: "prnReason", required: false, type: "text", maxLength: 1000}
		];
		var frequencyModel = new ObjectModel({propertyDefinitions: frequencyProperties});

		var FrequencyEditor = declare([ObjectEditor], {
			ndocNumber: null,
			discipline: null,
			cert: null,
			certStart: new Date(),
			certEnd: new Date(),
			minEditDate: null,
			objectModel: frequencyModel,
			withinPer: "null",
			tooManyWeeks: false,
			lastSpans: [],

			_frequencyEditorStore: null,

			templateString:
				'<div class="${baseClass}" data-dojo-attach-point="containerNode" style="width:100%;height:100%">'+
					'<div style="display:inline-block; vertical-align: top;">'+
						'<div data-dojo-attach-point="frequencyCalendarContainerContainer">'+
							'<div data-dojo-attach-point="frequencyCalendarContainer" style="font-size: 1.3em"></div>'+
						'</div>'+
						'<div data-dojo-attach-point="frequencyWeekContainer" style="font-size: small"></div>'+
						'<hr>'+
						'<div data-dojo-attach-point="frequencyLabelContainer" style="font-size: small"></div>'+
					'</div>'+
					'<div style="display:inline-block;">'+
						'<div data-dojo-attach-point="frequencyItemContainer"></div>'+
						'<div class="content-area">'+
							'<div class="content-flex-container">'+
								'<div class="flex-field-container vertical-field-align">'+
									'<div class="field-name field-name-medium">'+
										'<label>PRN Visits</label>'+
									'</div>'+
									'<div class="field-value field-value-small">'+
										'<input data-dojo-attach-point="prnVisitsNode">'+
									'</div>'+
								'</div>'+
								'<div class="flex-field-container vertical-field-align">'+
									'<div class="field-name field-name-large">'+
										'<label>PRN Reason</label>'+
									'</div>'+
									'<div class="field-value field-value-grand">'+
										'<input data-dojo-attach-point="prnReasonNode">'+
									'</div>'+
								'</div>'+
							'</div>'+
						'</div>'+
					'</div>'+
				'</div>',

			postMixInProperties: function() {
				var t = this;
				t.inherited(arguments);

				t._frequencyEditorStore = new Observable(new Memory());
				//t._frequencyEditorStore.set("idProperty", "startDate");

				t.getPropertyModel("prnReason").getField = function(node) {
					return new Textarea(this.getFieldProperties(), node);
				};
			},

			buildRendering: function() {
				var t = this;
				t.inherited(arguments);
			},

			postCreate: function() {
				var t = this;
				t.inherited(arguments);

				t.own(t._frequencyEditorStore.query().observe(function(frequencyItemEditor, removedFrom, insertedInto) {

					if (removedFrom > -1) {
						frequencyItemEditor.destroyRecursive();
					} else if (insertedInto > -1) {
						t.frequencyItemContainer.appendChild(frequencyItemEditor.domNode);
					}

					t.highlightCalendar();
				}, true));

				t.own(aspect.after(t, "_setObjectAttr", function(object) {
					t.ndocNumber = object.ndocNumber;
					t.discipline = object.discipline;
					t.cert = object.cert;
					t.tooManyWeeks = false;

					// Clear any existing frequencies
					t._frequencyEditorStore.query().forEach(function(editor) {
						t._frequencyEditorStore.remove(editor.id);
					});

					t.updatePrnDisabled();

					t.populateFrequencyWeeks();
					t.certStart = new Date(t.cert.startDate);
					t.certEnd = new Date(t.cert.stopDate);

					if (t.frequencyCalendar) {
						t.frequencyCalendar.destroyRecursive();
					}
					t.frequencyCalendar = new FrequencyCalendar({
						certStart: t.certStart,
						certEnd: t.certEnd,
						currentFocus: t.certStart
					}, t.frequencyCalendarContainer);
					t.frequencyCalendarContainerContainer.appendChild(t.frequencyCalendarContainer); // FIXME - can frequencyCalendarContainer be added back to the dom after destroy()?

					// Clicking on a particular day will focus you on the first field of the visit frequency that contains that day
					t.own(aspect.after(t.frequencyCalendar, "_onDayClick", function() {
						var value = t.frequencyCalendar.get("value");
						t._frequencyEditorStore.query().forEach(function(editor) {
							if (editor.containsDate(value)) {
								editor.startDateField.focus();
							} else if (editor.breakContainsDate(value)) {
								editor.breakReasonField.focus();
							}
						});
					}));

					// The individual days will appear as bold text if a visit for the discipline being modified is scheduled for that day
					serverRequest("Patient.Frequency.Utility", "GetVisitDates", { ndocNumber: t.ndocNumber, discipline: t.discipline, startDate: t.cert.startDate, stopDate: t.cert.stopDate }).then(function(dates) {
						t.frequencyCalendar.boldDates(dates);
					});

					t.initialLoad = true;
					arrayUtils.forEach(object.frequencyItems, t.addFrequencyItem, t);
					t.initialLoad = false;
					t.getSpans();

					t.decorateEditors();
				}, true));

				t.own(aspect.after(t,"_setActionTypeAttr",function(actionType){
					t._set('actionType', actionType);
					t.updateActionTypeForFrequencyItemEditors();
				}, true));

				t.own(aspect.after(t,"_setMinEditDateAttr",function(minEditDate){
					t._set('minEditDate', minEditDate);
					t.updateActionTypeForFrequencyItemEditors();
					t.updatePrnDisabled();
				}, true));

				t.registerChangeLogic("prnVisits", function(value) {
					t.require("prnReason", (value > 0));
					if (!t.inObjectLoad) {
						t.decorateEditors();
					}
				});

				t.registerChangeLogic("prnReason", function(value) {
					t.require("prnVisits", (value && (value.length > 0)));
					if (!t.inObjectLoad) {
						t.decorateEditors();
					}
				});
			},

			updateActionTypeForFrequencyItemEditors: function() {
				var t = this;
				var actionType = t.get('actionType');
				var minEditDate = t.get('minEditDate');
				t._frequencyEditorStore.query().forEach(function(editor) {
					if (editor.object.get('stopDate') < minEditDate) {
						editor.set('actionType',"LimitedEdit");
					} else {
						editor.set('actionType',actionType);
					}
				});
				t.decorateEditors();
			},

			updatePrnDisabled: function() {
				var t = this;
				var disabled = false;

				// Always disabled for View action
				if (t.get("actionType") === "View") {
					disabled = true;
				}

				// If minimum edit date is later than the cert period, disable the PRN fields
				var cert = t.get("cert");
				if (cert) {
					var certStop = new Date(cert.stopDate);
					if (t.get("minEditDate") > certStop) {
						disabled = true;
					}
				}

				// If PRN is not allowed for this discipline, disable the PRN fields
				if (arrayUtils.indexOf(frequencySettings.prnVisitDisciplines, t.get("discipline")) === -1) {
					disabled = true;
				}

				t.disable("prnVisits", disabled);
				t.disable("prnReason", disabled);
			},

			addFrequencyItem: function(frequencyItem) {
				var t = this;
				var MAX_FREQUENCIES = 10;	// current back end supports 10 frequency lines per discipline per cert

				if (t._frequencyEditorStore.data.length < MAX_FREQUENCIES) {
					var frequencyItemEditor = new FrequencyItemEditor({
						id: frequencyItem.id || t._frequencyEditorStore.data.length + 1,
						actionType: t.actionType,
						frequencyEditor: t
					});
					frequencyItemEditor.set("object", new FrequencyItem(frequencyItem));
					t._frequencyEditorStore.put(frequencyItemEditor);
				}
			},

			highlightCalendar: function() {
				var t = this;
				t.frequencyCalendar.highlightDateRange(t.certStart, t.certEnd, "");

				t._frequencyEditorStore.query().forEach(function(editor) {
					if (editor.object.startDate && editor.object.stopDate) {
						var frequencyStart = locale.parse(editor.object.startDate, {selector: "date", datePattern: "yyyy-MM-dd"});
						var frequencyStop = locale.parse(editor.object.stopDate, {selector: "date", datePattern: "yyyy-MM-dd"});
						t.frequencyCalendar.highlightDateRange(frequencyStart, frequencyStop, editor.highlightColor);
					}

					if (editor.breakStart && editor.breakStop) {
						t.frequencyCalendar.highlightDateRange(editor.breakStart, editor.breakStop, "#FFCCCC");
					}
				});
			},

			decorateEditors: function() {
				var t = this;
				var previousEditor = null;

				t._frequencyEditorStore.query().forEach(function(editor) {
					domStyle.set(editor.breakReasonContainer, "display", "none");

					editor.breakStart = null;
					editor.breakStop = null;

					if (previousEditor) {
						var thisFrequencyStart = locale.parse(editor.object.startDate, {selector: "date", datePattern: "yyyy-MM-dd"});
						var lastFrequencyStop = locale.parse(previousEditor.object.stopDate, {selector: "date", datePattern: "yyyy-MM-dd"});
						var frequencyGap = date.difference(lastFrequencyStop, thisFrequencyStart, "day") - 1;
						editor.startDateField.constraints.min = date.add(lastFrequencyStop, "day", 1);

						if (frequencyGap > 0) {
							editor.breakStart = date.add(lastFrequencyStop, "day", 1);
							editor.breakStop = date.add(thisFrequencyStart, "day", -1);
							var label = locale.format(editor.breakStart, {selector: 'date'}) + " - " + locale.format(editor.breakStop, {selector: 'date'}) + " (" + frequencyGap + " " + (frequencyGap == 1 ? "day" : "days") + ")";
							editor.breakDurationLabel.innerHTML = label;
							domStyle.set(editor.breakReasonContainer, "display", "");
							editor.require("breakReason", (true && frequencySettings.breakReasonRequired));
							editor.disable("breakReason", false);
						} else {
							domStyle.set(editor.breakReasonContainer, "display", "none");
							editor.require("breakReason", false);
							editor.disable("breakReason", true);
							editor.object.set("breakReason", "");
						}
						editor.updateFieldPinking();
					}

					// Always show the break reason field if it has a value
					if (editor.object.get("breakReason")) {
						domStyle.set(editor.breakReasonContainer, "display", "");
					}

					editor.setRequired();
					editor.updateStatus();
					editor.validate();
					previousEditor = editor;
				});

				if (t.frequencyCalendar) {
					t.highlightCalendar();
				}
			},

			populateFrequencyWeeks: function() {
				var t = this;
				domConstruct.empty(t.frequencyWeekContainer);
				arrayUtils.forEach(t.cert.weeks, function(week) {
					domConstruct.create("div", {innerHTML: week.label}, t.frequencyWeekContainer);
				});
			},

			getSpans: function() {
				var t = this;

				var frequencies = [];
				var editorIndex = {};
				var editorArray = [];
				var editorsValid = true;

				t._frequencyEditorStore.query().forEach(function(editor) {
					frequencies.push(editor.object);
					editorIndex[editor.id] = null;
					editorArray.push(editor.id);
					editorsValid = editorsValid && editor.isValid();
				});

				var options = {
					ndocNumber: t.ndocNumber,
					cert: t.object.cert,
					frequencies: frequencies,
					previousSpans: t.lastSpans
				};
				return serverRequest("Patient.Frequency.Utility", "Calcspan", options).then(function(results) {
					// Only show message if the value has changed
					if (t.tooManyWeeks != results.tooManyWeeks) {
						t.tooManyWeeks = results.tooManyWeeks;
						if (results.tooManyWeeks) {
							actionShowMessage("Too many weeks/days for this cert period");
						}
					}

					arrayUtils.forEach(results.spans, function(result) {
						var editor = t._frequencyEditorStore.get(result.id);
						if (editor && editor.object.startDate === result.startDate) {
							var stopDate = locale.parse(result.stopDate, {selector: "date", datePattern: "yyyy-MM-dd"});
							var startDate = locale.parse(result.startDate, {selector: "date", datePattern: "yyyy-MM-dd"});
							editor.object.set("startDate", startDate);
							editor.object.set("stopDate", stopDate);
							editorIndex[result.id] = "found"; // The frequency item still exists
						} else {
							if (editor) {
								t._frequencyEditorStore.remove(editor.id);
							}
							var lastEditor = t.getLastEditor();
							var newFrequency = {
								startDate: result.startDate,
								stopDate: result.stopDate,
								id: result.id,
								withinPer: lastEditor.object.get("withinPer"),
								perTimeUnit: lastEditor.object.get("perTimeUnit"),
								withinTimeUnit: lastEditor.object.get("withinTimeUnit")
							};
							t.addFrequencyItem(newFrequency);
							editorIndex[result.id] = "found";
						}
					});

					// Remove editors for frequency items that Calcspan killed, unless they are not at the end
					for (var iEditor = editorArray.length - 1; iEditor >= 0; iEditor--) {
						var editorId = editorArray[iEditor];
						if (editorIndex[editorId] == "found") {
							// we found one, no need to keep looking through the rest
							iEditor = 0;
						} else {
							t._frequencyEditorStore.remove(editorId);
						}
					}

					// Update lastEditor properties
					t.getLastEditor();

					t.decorateEditors();

					t.updateActionTypeForFrequencyItemEditors();
					t.updatePrnDisabled();

				}, function(error) {
					console.log(error);
				});
			},

			getLastEditor: function() {
				var t = this;
				var firstEditor = null;
				var previousEditor = null;
				t._frequencyEditorStore.query().forEach(function (editor) {
					editor.set("lastEditor", false);
					firstEditor = firstEditor || editor;
					previousEditor = editor;
				});
				if (firstEditor && previousEditor && firstEditor.id != previousEditor.id) {
					previousEditor.set("lastEditor", true);
				}
				return previousEditor;
			},

			retrieveFrequency: function(ndocNumber, discipline, cert) {
				return serverRequest("Patient.Frequency.Utility", "Query", {ndocNumber: ndocNumber, discipline: discipline, cert: cert});
			},

			retrieveIncompleteFrequency: function(ndocNumber, discipline, cert, orderId) {
				return serverRequest("Patient.Frequency.Utility", "GetIncompleteFrequency", {ndocNumber: ndocNumber,  discipline: discipline, cert: cert, orderId: orderId});
			},

			setFrequency: function(ndocNumber, discipline, cert, orderId, incomplete, reload) {
				var t = this;
				
				var promise;
				if (reload) {
					var newObject = lang.mixin({},t.get("object"));
					newObject = lang.mixin(newObject,t.get("value"));
					newObject.frequencies = newObject.frequencyItems;
					if (newObject.frequencies.length > 0) {
						newObject.frequencies[newObject.frequencies.length - 1].stopDate = locale.format(new Date(cert.stopDate),{selector:"date",datePattern:"yyyy-MM-dd"});
					}
					t.clearFrequency();
					
					var deferred = new Deferred();
					promise = deferred.promise;
					deferred.resolve(newObject);
				} else if (incomplete && orderId) {
					promise = t.retrieveIncompleteFrequency(ndocNumber, discipline, cert.id, orderId);
				} else {
					promise = t.retrieveFrequency(ndocNumber, discipline, cert.id);
				}

				return promise.then(function(results) {
					t.lastSpans = [];

					var lastFrequency = results.frequencies[results.frequencies.length - 1];
					var lastFrequencyStop = null;
					if (lastFrequency) {
						lastFrequencyStop = locale.parse(lastFrequency.stopDate, { selector: "date", datePattern: "yyyy-MM-dd" });
					}
					var certStart = new Date(cert.startDate);
					var certStop = new Date(cert.stopDate);

					// Add empty editor
					if (results.frequencies.length == 0) {
						var frequencyStart = certStart;
						var frequencyStop = certStop;
						if (t.get("minEditDate") > certStart) {
							frequencyStart = t.get("minEditDate");
						}
						if (frequencyStart > frequencyStop) {
							frequencyStop = frequencyStart;
						}
						results.frequencies.push({
							startDate: locale.format(frequencyStart, {
								selector: 'date',
								datePattern: "yyyy-MM-dd"
							}),
							stopDate: locale.format(frequencyStop, {
								selector: 'date',
								datePattern: "yyyy-MM-dd"
							})
						});
					} else if (lastFrequency && lastFrequencyStop < certStop) {
						// Add an editor if there are unallocated days left at the end of the cert
						results.frequencies.push({
							startDate: locale.format(date.add(lastFrequencyStop, "day", 1), {
								selector: 'date',
								datePattern: "yyyy-MM-dd"
							}),
							stopDate: locale.format(certStop, {
								selector: 'date',
								datePattern: "yyyy-MM-dd"
							})
						});
					}

					var Frequency = t.objectModel.getObject();
					var frequency = new Frequency({
						ndocNumber: ndocNumber,
						discipline: discipline,
						cert: cert,
						frequencyItems: results.frequencies,
						prnVisits: results.prnVisits,
						prnReason: results.prnReason
					});
					t.set("object", frequency);
				}, actionShowMessage);
			},

			clearFrequency: function() {
				var t = this;
				t._frequencyEditorStore.query().forEach(function(editor) {
					t._frequencyEditorStore.remove(editor.id);
				});
				if (t.object) {
					t.object.set("cert", null);
					t.object.set("prnVisits", "");
					t.object.set("prnReason", "");
				}
				t.cert = null;
				t.tooManyWeeks = false;
			},

			isValid: function() {
				var t = this;

				for (var property in t.objectModel.propertyDefinitions) {
					var fieldName = t.objectModel.propertyDefinitions[property].name;

					var widget = t[fieldName+"Field"];
					if (widget && (typeof widget.isValid !== "undefined") && !widget.isValid()) {
						return false;
					}
				}

				var frequencyLinesValid = true;
				var frequencyLineCount = 0;

				// Make sure all frequency lines are valid, except the "last" one (first is not also considered last)
				t._frequencyEditorStore.query().forEach(function(editor) {
					++frequencyLineCount;
					frequencyLinesValid = frequencyLinesValid && (editor.get("status") == "complete" || editor.get("lastEditor"));
				});

				var hasPRN = false;
				if ((t.object.get("prnVisits") > 0) && (t.object.get("prnReason") != "")) {
					hasPRN = true;
				}

				var valid = false;
				if (frequencyLinesValid) {
					valid = true;
				} else if (hasPRN && (frequencyLineCount === 1)) {
					valid = true;
				}

				return valid;
			},

			_setDisabledAttr: function(value) {
				arrayUtils.forEach(t.objectModel.getPropertyModels(), function(propertyModel) {
					propertyModel.fieldProperties.disabled = true;
				});
			},

			_getValueAttr: function() {
				// summary:
				//		Support get('value')

				var t = this;
				var value = {};

				if (t.object) {
					value = t.object.getData();
					value.frequencyItems = [];
					t._frequencyEditorStore.query().forEach(function(editor) {
						value.frequencyItems.push(editor.object.getData());
					});
				}

				return value;
			}
		});

		return FrequencyEditor;
	}
);
